{
    "links": [
        {
            "href": "%(host)s/v3/89afd400-b646-4bbc-b12b-c0a4d63e5bd3/qos-specs/%(uuid)s",
            "rel": "self"
        },
        {
            "href": "%(host)s/89afd400-b646-4bbc-b12b-c0a4d63e5bd3/qos-specs/%(uuid)s",
            "rel": "bookmark"
        }
    ],
    "qos_specs": {
        "consumer": "back-end",
        "id": "%(uuid)s",
        "name": "reliability-spec",
        "specs": {}
    }
}