#  Copyright (C) 2021-2022 YADRO.
#  All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

HOST_GROUPS = 'personalities/v1/config/groups'
HOSTS = 'personalities/v1/config/hosts'
RESOURCE = 'personalities/v1/personalities/block/%s'
RESOURCE_DETAIL = 'personalities/v1/personalities?id=%s'
RESOURCE_HEALTH = 'health/v1/personalities?id=%s'
RESOURCE_MAPPING = 'personalities/v1/config/resource_mapping'
VOLUME_TO_HOST = 'personalities/v1/personalities/block/%s/hosts/%s'
ALL_RESOURCES = 'personalities/v1/personalities'
POOLS = 'health/v1/pools'
STATISTICS = 'health/v1/statistics/current'
IP_PORTS = 'osmgr/v1/ports/%s'
RESOURCE_COUNT = 'personalities/v1/personalities/block/countPerPool'
TATLIN_VERSION = 'upmgr/v1/version'
