{
    "volumes": [
        {
            "id": "%(uuid)s",
            "links": [
                {
                    "href": "%(host)s/v3/89afd400-b646-4bbc-b12b-c0a4d63e5bd3/volumes/%(uuid)s",
                    "rel": "self"
                },
                {
                    "href": "%(host)s/89afd400-b646-4bbc-b12b-c0a4d63e5bd3/volumes/%(uuid)s",
                    "rel": "bookmark"
                }
            ],
            "name": null
        }
    ]
}