# Stubs for oslo_i18n._translate (Python 3)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional

def translate(obj: Any, desired_locale: Optional[Any] = ...): ...
