{
    "volume_type": {
        "name": "%(name)s",
        "description": "%(description)s",
        "os-volume-type-access:is_public": "%(bool)s",
        "extra_specs": {
            "capabilities": "gpu"
        }
    }
}
