{
    "os-volume_upload_image": {
        "container_format": "bare",
        "disk_format": "raw",
        "display_description": null,
        "id": "3a81fdac-e8ae-4e61-b6a2-2e14ff316f19",
        "image_id": "de75b74e-7f0d-4b59-a263-bd87bfc313bd",
        "image_name": "test",
        "protected": false,
        "size": 1,
        "status": "uploading",
        "updated_at": "2017-06-05T08:44:28.000000",
        "visibility": "private",
        "volume_type": null
    }
}