# Stubs for oslo_i18n (Python 3)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from ._factory import *
from ._gettextutils import *
from ._lazy import *
from ._translate import *
